﻿// demo of multi-swarm optimization using C#
// solve Rastigin's function

using System;
namespace MultiSwarm
{
  class MultiSwarmProgram
  {
    static void Main(string[] args)
    {
      Console.WriteLine("\nBegin multi-swarm optimization demo\n");

      int dim = 6;
      Console.WriteLine("Goal is to solve Rastrigin's function of " + dim + " variables");
      Console.Write("Function has known min value of 0.0 at (");
      for (int i = 0; i < dim -1; ++i)
        Console.Write("0, ");
      Console.WriteLine("0)");

      int numSwarms = 4;
      int numParticles = 5;
      int maxEpochs = 2000;
      double minX = -10.0;
      double maxX = 10.0;

      Console.WriteLine("\nSetting numSwarms = " + numSwarms);
      Console.WriteLine("Setting numParticles = " + numParticles);
      Console.WriteLine("Setting maxEpochs = " + maxEpochs);
 
      Console.WriteLine("\nStart MSO solve\n");
      double[] bestPosition = Solve(dim, numSwarms, numParticles, maxEpochs, minX, maxX);
      double bestError = Error(bestPosition);
      Console.WriteLine("\nDone");

      Console.WriteLine("\nBest solution found: ");
      ShowVector(bestPosition, 4, true);
      Console.WriteLine("Error of best position = " + bestError.ToString("F8"));

      Console.WriteLine("\nEnd multi-swarm optimization demo\n");
      Console.ReadLine();
    } // Main

    static void ShowVector(double[] vector, int decimals, bool newLine)
    {
      for (int i = 0; i < vector.Length; ++i)
        Console.Write(vector[i].ToString("F" + decimals) + " ");
      Console.WriteLine("");
      if (newLine == true) Console.WriteLine("");
    }

    static double[] Solve(int dim, int numSwarms, int numParticles, int maxEpochs, double minX, double maxX)
    {
      Random rnd = new Random(0);
      Swarm[] swarms = new Swarm[numSwarms];
      for (int i = 0; i < numSwarms; ++i)
        swarms[i] = new Swarm(numParticles, dim, minX, maxX); // random particles in each swarm
      double[] bestMultiPosition = new double[dim];
      double bestMultiError = double.MaxValue;
      for (int i = 0; i < numSwarms; ++i)
      {
        if (swarms[i].bestSwarmError < bestMultiError)
        {
          bestMultiError = swarms[i].bestSwarmError;
          Array.Copy(swarms[i].bestSwarmPosition, bestMultiPosition, dim);
        }
      }

      int epoch = 0;
      double w = 0.729; // inertia
      double c1 = 1.49445; // particle / cognitive
      double c2 = 1.49445; // swarm / social
      double c3 = 0.3645; // multiswarm / global
      double pDeath = 0.005;      // prob of particle death
      double pImmigrate = 0.005;  // prob of particle immigration
      //double minV = 0.1 * minX;   // typically 0.1 * -10 = -1.0
      //double maxV = 0.1 * maxX;   // typically +1.0

      //int[] sequence = new int[numParticles]; // process particles in random order
      //for (int i = 0; i < sequence.Length; ++i)
      //  sequence[i] = i;

      while (epoch < maxEpochs)
      {
        ++epoch;

        if (epoch % 100 == 0 && epoch < maxEpochs)
        {
          Console.WriteLine("Epoch = " + epoch + " best error = " + bestMultiError.ToString("F4"));
        }

        for (int i = 0; i < numSwarms; ++i) // each swarm
        {
          // Shuffle(sequence, rnd); // move particles in random sequence
          for (int j = 0; j < numParticles; ++j) // each particle
          {
            //int j = sequence[pj];
            double p1 = rnd.NextDouble(); 
            if (p1 < pDeath)
            {
              swarms[i].particles[j] = new Particle(dim, minX, maxX); // new random position
              if (swarms[i].particles[j].error < swarms[i].bestSwarmError) // new swarm best by luck?
              {
                swarms[i].bestSwarmError = swarms[i].particles[j].error;
                Array.Copy(swarms[i].particles[j].position, swarms[i].bestSwarmPosition, dim);
                if (swarms[i].particles[j].error < bestMultiError) // if a new swarm best, maybe also a new global best?
                {
                  bestMultiError = swarms[i].particles[j].error;
                  Array.Copy(swarms[i].particles[j].position, bestMultiPosition, dim);
                }
              }
            }
            // an alternative is to maintain a particle age and die with high prob after a certain age reached
            // another option is to maintain particle health/weakness (related to either ratio of times improved / loop count
            // or number consecutive improves or consecutive non-improves) and die with high prob when health is low 

            double p2 = rnd.NextDouble();
            if (p2 < pImmigrate)
            {
              int otherSwarm = rnd.Next(0, numSwarms);
              int otherParticle = rnd.Next(0, numParticles);
              Particle tmp = swarms[i].particles[j];
              swarms[i].particles[j] = swarms[otherSwarm].particles[otherParticle];
              swarms[otherSwarm].particles[otherParticle] = tmp;

              if (swarms[i].particles[j].error < swarms[otherSwarm].bestSwarmError) // new (other) swarm best?
              {
                swarms[otherSwarm].bestSwarmError = swarms[i].particles[j].error;
                Array.Copy(swarms[i].particles[j].position, swarms[otherSwarm].bestSwarmPosition, dim);
              }
              if (swarms[otherSwarm].particles[otherParticle].error < swarms[i].bestSwarmError) // new (curr) swarm best?
              {
                swarms[i].bestSwarmError = swarms[otherSwarm].particles[otherParticle].error;
                Array.Copy(swarms[otherSwarm].particles[otherParticle].position, swarms[i].bestSwarmPosition, dim);
              }
              // not possible for a new global best
            }

            for (int k = 0; k < dim; ++k) // update velocity. each x position component
            {
              double r1 = rnd.NextDouble();
              double r2 = rnd.NextDouble();
              double r3 = rnd.NextDouble();

              swarms[i].particles[j].velocity[k] = (w * swarms[i].particles[j].velocity[k]) +
                (c1 * r1 * (swarms[i].particles[j].bestPartPosition[k] - swarms[i].particles[j].position[k])) +
                (c2 * r2 * (swarms[i].bestSwarmPosition[k] - swarms[i].particles[j].position[k])) +
                (c3 * r3 * (bestMultiPosition[k] - swarms[i].particles[j].position[k]));

              if (swarms[i].particles[j].velocity[k] < minX) // constrain velocities
                swarms[i].particles[j].velocity[k] = minX;
              else if (swarms[i].particles[j].velocity[k] > maxX)
                swarms[i].particles[j].velocity[k] = maxX;
            }

            for (int k = 0; k < dim; ++k) // update position
            {
              swarms[i].particles[j].position[k] += swarms[i].particles[j].velocity[k];
              // constrain all xi
              if (swarms[i].particles[j].position[k] < minX)
                //swarms[i].particles[j].position[k] = (maxX - minX) * rnd.NextDouble() + minX;
                swarms[i].particles[j].position[k] = minX;
              else if (swarms[i].particles[j].position[k] > maxX)
                //swarms[i].particles[j].position[k] = (maxX - minX) * rnd.NextDouble() + minX;
                swarms[i].particles[j].position[k] = maxX;
            }

            // update error
            swarms[i].particles[j].error = MultiSwarmProgram.Error(swarms[i].particles[j].position);

            // check if new best error for this particle
            if (swarms[i].particles[j].error < swarms[i].particles[j].bestPartError)
            {
              swarms[i].particles[j].bestPartError = swarms[i].particles[j].error;
              Array.Copy(swarms[i].particles[j].position, swarms[i].particles[j].bestPartPosition, dim);
            }

            if (swarms[i].particles[j].error < swarms[i].bestSwarmError) // new swarm best?
            {
              swarms[i].bestSwarmError = swarms[i].particles[j].error;
              Array.Copy(swarms[i].particles[j].position, swarms[i].bestSwarmPosition, dim);
            }

            if (swarms[i].particles[j].error < bestMultiError) // new global best?
            {
              bestMultiError = swarms[i].particles[j].error;
              Array.Copy(swarms[i].particles[j].position, bestMultiPosition, dim);
            }
          } // each particle
        } // each swarm
      } // while
      return bestMultiPosition;
    } // Solve

    private static void Shuffle(int[] sequence, Random rnd)
    {
      for (int i = 0; i < sequence.Length; ++i)
      {
        int r = rnd.Next(i, sequence.Length);
        int tmp = sequence[r];
        sequence[r] = sequence[i];
        sequence[i] = tmp;
      }
    }

    public static double Error(double[] position)
    {
      double trueMin = 0.0;
      double valAtPosition = 0.0;
      for (int i = 0; i < position.Length; ++i)
      {
        double xi = position[i];
        valAtPosition += (xi * xi) - (10 * Math.Cos(2 * Math.PI * xi)) + 10;
      }
      return Math.Abs(trueMin - valAtPosition);
    }
  } // Program class

  public class Particle
  {
    static Random ran = new Random(0);
    public double[] position;
    public double[] velocity;
    public double error;
    public double[] bestPartPosition;
    public double bestPartError;

    public Particle(int dim, double minX, double maxX) // random position
    {
      position = new double[dim];
      velocity = new double[dim];
      bestPartPosition = new double[dim];
      for (int i = 0; i < dim; ++i)
      {
        position[i] = (maxX - minX) * ran.NextDouble() + minX;
        velocity[i] = (maxX - minX) * ran.NextDouble() + minX;
      }
      error = MultiSwarmProgram.Error(position);
      bestPartError = error;
      Array.Copy(position, bestPartPosition, dim);
    }
  } // Particle

  public class Swarm
  {
    public Particle[] particles;
    public double[] bestSwarmPosition;
    public double bestSwarmError;

    public Swarm(int numParticles, int dim, double minX, double maxX)
    {
      bestSwarmError = double.MaxValue;
      bestSwarmPosition = new double[dim];
      particles = new Particle[numParticles];
      for (int i = 0; i < particles.Length; ++i)
      {
        particles[i] = new Particle(dim, minX, maxX); // random position
        if (particles[i].error < bestSwarmError)
        {
          bestSwarmError = particles[i].error;
          Array.Copy(particles[i].position, bestSwarmPosition, dim);
        }
      }
    } // ctor
   } // Swarm
} // ns
